'''
Dynamic Time Warping based Hierarchical Agglomerative Clustering
================================================================
Codes to perform Dynamic Time Warping Based Hierarchical Agglomerative Clustering of GPS data

:author: Utpal Kumar
:date: 2021/08
:copyright: 2021, Institute of Earth Sciences, Academia Sinica.
'''
from analysis_support import dec2dt, dec2dt_scalar, toYearFraction
from dtw_analysis import dtw_signal_pairs, plot_signals, plot_cluster, shuffle_signals, noise_robustness_test, dtw_clustering
from leastSquareModeling import LSQmodules, lsqmodeling
from plot_linear_trend import compute_interpolation, plot_linear_trend_on_map
from plot_stations import plot_station_map

# version
__version__ = "1.0.9"
__author__ = "Utpal Kumar"
__copyright__ = "Copyright 2021 Utpal Kumar"
__license__ = "Apache License, Version 2.0"
